#pragma once

#include <utility>

namespace detail
{

    template <class T>
    T replace(T& dst, T src = T())
    {
        using std::swap;
        swap(dst, src);
        return src;
    }

}

template <class T>
class Uniq_ptr
{
    T* raw_;

public:
    using element_type = T;
    using pointer      = element_type*;
    using reference    = element_type&;

    Uniq_ptr() noexcept
            : raw_(nullptr)
    { }

    Uniq_ptr(std::nullptr_t) noexcept
            : Uniq_ptr()
    { }

    explicit Uniq_ptr(pointer raw) noexcept
            : raw_(raw)
    { }

    Uniq_ptr(Uniq_ptr&& that) noexcept
            : raw_(detail::replace(that.raw_))
    { }

    Uniq_ptr& operator=(Uniq_ptr&& that) noexcept
    {
        Uniq_ptr temp(std::move(that));
        swap(temp);
        return *this;
    }

    Uniq_ptr(Uniq_ptr const&) = delete;
    Uniq_ptr& operator=(Uniq_ptr const&) = delete;

    ~Uniq_ptr()
    {
        delete raw_;
    }

    explicit operator bool() const noexcept
    {
        return bool(raw_);
    }

    pointer get() const noexcept
    {
        return raw_;
    }

    reference operator*() const
    {
        return *raw_;
    }

    pointer operator->() const noexcept
    {
        return raw_;
    }

    void reset(pointer raw = pointer()) noexcept
    {
        *this = Uniq_ptr(raw);
    }

    pointer release() noexcept
    {
        return detail::replace(raw_);
    }

    void swap(Uniq_ptr& other) noexcept
    {
        std::swap(raw_, other.raw_);
    }

    Uniq_ptr clone() const
    {
        return Uniq_ptr(new element_type(*raw_));
    }
};

template <class T, class...Args>
Uniq_ptr<T> make_Uniq(Args&& ... args)
{
    return Uniq_ptr<T>(new T(std::forward<Args>(args)...));
}

template <class T>
void swap(Uniq_ptr<T>& a, Uniq_ptr<T>& b) noexcept
{
    a.swap(b);
}

// Comparisons between Uniq_ptrs, and between a Uniq_ptr and a
// nullptr:

template <class T1, class T2>
bool operator==(Uniq_ptr<T1> const& a, Uniq_ptr<T2> const& b)
{
    return a.get() == b.get();
}

template <class T1, class T2>
bool operator!=(Uniq_ptr<T1> const& a, Uniq_ptr<T2> const& b)
{
    return a.get() != b.get();
}

template <class T1>
bool operator==(Uniq_ptr<T1> const& a, std::nullptr_t)
{
    return a.get() == nullptr;
}

template <class T1>
bool operator!=(Uniq_ptr<T1> const& a, std::nullptr_t)
{
    return a.get() != nullptr;
}

template <class T2>
bool operator==(std::nullptr_t, Uniq_ptr<T2> const& b)
{
    return nullptr == b.get();
}

template <class T2>
bool operator!=(std::nullptr_t, Uniq_ptr<T2> const& b)
{
    return nullptr != b.get();
}
