#pragma once

#include <functional>
#include <iostream>
#include <memory>
#include <vector>

template <template <class> class Ptr, class State>
class Tracker
{
public:
    class Guard;

    using state_t = State;
    using ptr_t   = Ptr<Guard>;

    explicit Tracker(state_t const& state = state_t());

    state_t const& get_state() const;

    template <class F>
    ptr_t delay(F);

    template <class T>
    ptr_t operator<<(T);

protected:
    using state_ptr_t = std::shared_ptr<state_t>;

    explicit Tracker(state_ptr_t const&) noexcept;

    state_ptr_t state_;
};

template <template <class> class Ptr, class State>
class Tracker<Ptr, State>::Guard : public Tracker
{
public:
    Guard(Guard const&) = delete;
    Guard& operator=(Guard const&) = delete;

    ~Guard();

private:
    friend Tracker;

    using fn_t = std::function<void(state_t&)>;

    template <class F>
    Guard(state_ptr_t const&, F);

    fn_t fn_;
};

//template <template <class> class Ptr, class T>
//class log_tracker

template<template <class> class Ptr, class State>
Tracker<Ptr, State>::Tracker(state_t const& state)
        : state_(new state_t(state))
{ }

template <template <class> class Ptr, class State>
auto Tracker<Ptr, State>::get_state() const -> state_t const&
{
    return *state_;
}

template <template <class> class Ptr, class State>
template <class F>
auto Tracker<Ptr, State>::delay(F fn) -> ptr_t
{
    return ptr_t(new Guard(state_, std::move(fn)));
}

template<template <class> class Ptr, class State>
template<class T>
auto Tracker<Ptr, State>::operator<<(T element) -> ptr_t
{
    return delay([=](state_t& state) {
        state.push_back(std::move(element));
    });
}

template<template <class> class Ptr, class State>
Tracker<Ptr, State>::Tracker(state_ptr_t const& state) noexcept
        : state_(state)
{ }

template <template <class> class Ptr, class State>
template <class F>
Tracker<Ptr, State>::Guard::Guard(state_ptr_t const& state, F fn)
        : Tracker(state)
        , fn_(std::move(fn))
{ }

template <template <class> class Ptr, class State>
Tracker<Ptr, State>::Guard::~Guard()
{
    try {
        fn_(*state_);
    } catch (...) {
        std::cerr << "*** guard threw in destructor ***\n";
    }
}


